/*
 * b16 core: 16 bits, 
 * inspired by c18 core from Chuck Moore
 *
 * Instruction set:
 * 1, 5, 5, 5 bits
 *     0    1    2    3    4    5    6    7
 *  0: nop  call jmp  ret  jz   jnz  jc   jnc
 *  /3      exec goto ret  gz   gnz  gc   gnc
 *  8: xor  com  and  or   +    +c   u2/  c2/
 * 10: !+   @+   @    lit  c!+  c@+  c@   litc
 *  /1 !.   @.   @    lit  c!.  c@.  c@   litc
 * 18: nip  drop over dup  >r        r>
 */
 
`define L [l-1:0]
`define DROP { sp, T } <= { spinc, N } 
`timescale 1ns / 1ns

module alu(res, carry, zero, T, N, c, inst);
   parameter l=16;
   input `L T, N;
   input c;
   input [2:0] inst;
   output `L res;
   output carry, zero;

   wire prop, andor, selr;

   assign #1 { prop, selr, andor } = inst;

   wire        `L r1, r2;
   wire [l:0]  carries;

   assign #1 r1 = T ^ N ^ carries;
   assign #1 r2 = (T & N) | 
                  (T & carries`L) | 
                  (N & carries`L);
   assign #1 carries = 
      prop ? { r2[l-1:0], (c | selr) & andor } 
           : { c, {(l){andor}}};
   assign #1 res = (selr & ~prop) ? r2 : r1;
   assign #1 carry = carries[l];
   assign #1 zero = ~|T;
endmodule // alu
module stack(clk, sp, spdec, push, in, out);
   parameter dep=2, l=16;
   input clk, push;
   input [dep-1:0] sp, spdec; 
   input `L in;
   output `L out; 
`ifdef SYNCSTACK
// on FPGAs, SYNCSTACK is preferred
   reg `L stackmem[0:(1<<dep)-1];
   always @(posedge clk)
     if(push) 
       stackmem[spdec] <= in;

   assign out = stackmen[sp];
`else 
   wire [dep-1:0] #1 spset = spdec;
   wire #1 write = push & ~clk;
   wire `L #1 ind = in;
   reg `L stackmem[0:(1<<dep)-1];

   always @(write or spset or ind)
     if(write) stackmem[spset] <= ind;

   assign #1 out = stackmem[sp]; 
`endif
endmodule // stack
module cpu(clk, run, reset, addr, rd, wr, 
           data, dataout, READY);
   parameter show=0, l=16, sdep=2, rdep=2;
   input clk, run, reset, READY;
   output `L addr;
   output rd;
   output [1:0] wr;
   input  `L data;
   output `L dataout;
   reg [sdep-1:0] sp;
   reg [rdep-1:0] rp;

   reg `L T, I, P;

   reg [1:0] state;
   reg c;
   // instruction and branch target selection   
   reg [4:0] inst;
   reg `L jmp;

   always @(state or I or data)
      case(state[1:0])
        2'b00: inst <= { 4'b0, data[15] };
        2'b01: inst <=   I[14:10];
        2'b10: inst <=   I[9:5];
        2'b11: inst <=   I[4:0];
      endcase // casez(state)

   always @(state or I or P or T or data)
      case(state[1:0])
        2'b00: jmp <= { data[14:0], 1'b0 };
        2'b01: jmp <= { P[15:11], I[9:0], 1'b0 };
        2'b10: jmp <= { P[15:6], I[4:0], 1'b0 };
        2'b11: jmp <= { T[15:1], 1'b0 };
      endcase // casez(state)
   wire `L res, toN, N;
   wire carry, zero;

   alu #(l) alu16(res, carry, zero, 
                  T, N, c, inst[2:0]);
   wire `L incaddr, toR, R, dataw;
   wire tos2r, tos2n;
   wire incby, addrsel, access, rd;
   wire [1:0] wr;

   assign incby = (inst[4:2] != 3'b101);
   assign access = (inst[4:3]==2'b10);
   assign addrsel = rd ? (access & (~&inst[1:0])) 
                       : |wr;
   assign rd = (state==2'b00) || 
               (access && (|inst[1:0]));
   assign wr = (access && (~|inst[1:0])) ?
               { ~inst[2] | ~T[0], 
                 ~inst[2] | T[0] } : 2'b00;
   assign tos2r = (inst == 5'b11100);
   assign tos2n = (!rd | (inst[1:0] == 2'b11));
   assign addr = addrsel ? T : P;
   assign incaddr = 
       { addr[l-1:1] + (incby | addr[0]), 
                      ~(incby | addr[0]) };
   assign toR = tos2r ? T :
                { |state ? P[15:1] 
                         : incaddr[15:1], c });
   assign toN = tos2n ? T : dataw;
   assign dataw = incby ? data :
                  { 8'h00, addr[0] ? data[7:0] 
                                   : data[l-1:8] }; 
   assign dataout = { incby ? N[15:8] 
                            : N[7:0], N[7:0] }; 
   reg dpush, rpush;

   always @(state or inst or rd or READY)
     begin
        rpush <= 1'b0;
        dpush <= |state[1:0] & rd;
        casez(inst)
           5'b00001: rpush <= |state[1:0] | READY;
           5'b11100: rpush <= 1'b1;
           5'b11?1?: dpush <= 1'b1;
        endcase // case(inst)
     end
   wire [sdep-1:0] spdec, spinc;
   wire [rdep-1:0] rpdec, rpinc;

   stack #(sdep,l) dstack(clk, sp, spdec, 
                          dpush, toN, N);
   stack #(rdep,l) rstack(clk, rp, rpdec, 
                          rpush, toR, R);

   assign spdec = sp-{{(sdep-1){1'b0}}, 1'b1};
   assign spinc = sp+{{(sdep-1){1'b0}}, 1'b1};
   assign rpdec = rp-{{(rdep-1){1'b0}}, 1'b1};
   assign rpinc = rp+{{(rdep-1){1'b0}}, 1'b1};
   reg [1:0] nextstate;

   always @(inst or state)
      casez(inst) 
         5'b00000: nextstate <= state[1:0] + 2'b01; 
         5'b00???: nextstate <= 2'b00;
         5'b?????: nextstate <= state[1:0] + 2'b01; 
      endcase // casez(inst[0:2])

   always @(posedge clk or negedge reset)
      if(!reset) begin
         state <= 2'b11;
         P <= 16'h07FE;
         T <= 16'h0000;
         I <= 16'h0000;
         c <= 1'b0;
         sp <= 0;
         rp <= 0;
      end else if(run) begin
         if(show) begin
            $write("%b[%b] T=%b%x:%x[%x], ",
                   inst, state, c, T, N, sp);
            $write("P=%x, I=%x, R=%x[%x], res=%b%x\n",
                   P, I, R, rp, carry, res);
         end
         if(~|state[1:0] || 
            ((inst[4:3] == 2'b10) && (&inst[1:0]))) 
            P <= incaddr;
         if(|state[1:0]) begin 
            if(rd && { inst[4:3], inst[1:0] } != 4'b1010) 
               sp <= spdec;
            if(|wr) sp <= spinc;
         end else begin 
            I <= data; 
            if(!data[15]) state[1:0] <= 2'b01;
         end
         state <= nextstate;
         casez(inst)
            5'b00001: begin
               rp <= rpdec;
               P <= jmp;
               c <= 1'b0;
               if(state == 2'b11) `DROP;
            end // case: 5'b00001
            5'b00010: begin
               P <= jmp;
               if(state == 2'b11) `DROP;
            end
            5'b00011: { rp, c, P } <= 
                      { rpinc, R[0], R[l-1:1], 1'b0 };
            5'b001??: begin
               if((inst[1] ? c : zero) ^ inst[0]) 
                  P <= jmp;
               `DROP;
            end
            5'b01001: { c, T } <= { 1'b1, ~T };
            5'b0111?: { T, c } <= { inst[0] ? c : T[15], T };
            5'b01???: { sp, c, T } <= { spinc, carry, res }; 
            5'b10?0?: begin
               if(nextstate != 2'b10) T <= incaddr;
               sp <= rd ? spdec : spinc;
            end
            5'b10?1?: T <= dataw;
            5'b11000: sp <= spinc;
            5'b11001: `DROP;
            5'b11010: { sp, T } <= { spdec, N };
            5'b11011: sp <= spdec;
            5'b11100: begin
               rp <= rpdec; `DROP;
            end // case: 5'b11100
            5'b11110: begin
               { sp, T } <= { spdec, R };
               rp <= rpinc;
            end // case: 5'b11110
         endcase // case(inst)
      end // else: !if(reset)

endmodule // cpu
